/********************************************************************
 K Win - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2011 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#ifndef GENERIC_ANIMATIONS_H
#define GENERIC_ANIMATIONS_H

#if HAVE_KWIN_4_8
#include <kwinanimationeffect.h>
#else
#include "animationeffect.h"
#endif

#include <QHash>

#define OLD_SIGNATURE 0

namespace KWin
{

class GenericAnimations : public AnimationEffect
{
    Q_OBJECT
public:
    GenericAnimations();
    ~GenericAnimations();
    void reconfigure( ReconfigureFlags );
    void trigger(QString animation, WId id);

protected:
    void animationEnded( EffectWindow *, Attribute, uint );
    void timerEvent(QTimerEvent *te);

private slots:
#if OLD_SIGNATURE // NOTICE moc cannot handle preprocessor instructions, so you have to un/comment this by hand :-(
/*
    void desktopChanged(int oldDesktop, int newDesktop);
    void moveResizeStarted(EffectWindow *w);
    void moveResizeStopped(EffectWindow *w);
    void opacityChanged( EffectWindow *w, qreal oldOpacity, qreal newOpacity );
//     void propertyChanged(EffectWindow *w, long atom);
    void windowActivated( EffectWindow *w );
    void windowAdded( EffectWindow *w );
    void windowClosed( EffectWindow *w );
    void windowDeleted( EffectWindow *w );
    void windowGeometryShapeChanged( EffectWindow *w, const QRect &old );
    void windowMinimized( EffectWindow *w );
    void windowUnminimized( EffectWindow *w );
*/
#else
    void desktopChanged(int oldDesktop, int newDesktop);
    void moveResizeStarted(KWin::EffectWindow *w);
    void moveResizeStopped(KWin::EffectWindow *w);
    void opacityChanged(KWin::EffectWindow *w, qreal oldOpacity, qreal newOpacity );
//     void propertyChanged(EffectWindow *w, long atom);
    void windowActivated(KWin::EffectWindow *w );
    void windowAdded(KWin::EffectWindow *w );
    void windowClosed(KWin::EffectWindow *w );
    void windowDeleted(KWin::EffectWindow *w );
    void windowGeometryShapeChanged(KWin::EffectWindow *w, const QRect &old );
    void windowMinimized(KWin::EffectWindow *w );
    void windowUnminimized(KWin::EffectWindow *w );
#endif

    void desktopChanged(int oldDesktop, int newDesktop,  KWin::EffectWindow *w );

private:
    void animateResize( EffectWindow *w, const QRect &from, const QRect &to );
    bool slideInOutPopup(EffectWindow *w, bool in);

private:
    typedef QHash<QString, QList<AniData> > AniHash;
    AniHash m_animations;
    typedef QPair<QRect, QRect> PendingResize;
    typedef QList<PendingResize> PendingResizeList;
    typedef QMap<EffectWindow*, PendingResizeList> PendingResizes;
    PendingResizes m_pendingResizes;
    EffectWindow *m_lastActive, *m_moveResizeWindow, *m_implicitActivationCandidate;
    bool m_animateMaximize, m_resizeConnected, m_slidePopups;
    int m_implicitActivationResetTimer, m_activationTimer;
    EffectWindowList m_slidingPopups;
};

} //namespace

#endif // GENERIC_ANIMATIONS_H
