/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2009 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "generic_animations_config.h"
#include "anidata_p.h"
#include <kwineffects.h>
#include <netwm_def.h>

#include <klocale.h>
#include <KActionCollection>
#include <kaction.h>
#include <kconfiggroup.h>
#include <KInputDialog>
#include <KMessageBox>

#include <QApplication>
#include <QDesktopWidget>
#include <QEasingCurve>
#include <QVBoxLayout>

#include <QtDebug>


using namespace KWin;

KWIN_EFFECT_CONFIG( generic_animations, GenericAnimationsConfig )

#include "animation.defaults"

GenericAnimationsConfigForm::GenericAnimationsConfigForm(QWidget* parent) : QWidget(parent)
{
    setupUi(this);
}

GenericAnimationsConfig::GenericAnimationsConfig(QWidget* parent, const QVariantList& args)
: KCModule(EffectFactory::componentData(), parent, args)
{
    initDefaultAnimations();

    attributeString[AnimationEffect::Opacity]    = i18n("Opacity");
    attributeString[AnimationEffect::Brightness] = i18n("Brightness");
    attributeString[AnimationEffect::Saturation] = i18n("Saturation");
    attributeString[AnimationEffect::Scale]      = i18n("Scale");
    attributeString[AnimationEffect::Translation]= i18n("Translation");
    attributeString[AnimationEffect::Rotation]   = i18n("Rotation");
    attributeString[AnimationEffect::Position]   = i18n("Position");
    attributeString[AnimationEffect::Size]       = i18n("Size");
    attributeString[AnimationEffect::Clip]       = i18n("Clipping");


    QVBoxLayout* layout = new QVBoxLayout(this);
    layout->addWidget(m_ui = new GenericAnimationsConfigForm(this));

    QFont fnt;
    fnt.setPointSize(qRound(fnt.pointSize()*15.0/10.0));
    fnt.setBold(true);
    m_ui->currentAnimation->setFont(fnt);
    fnt.setPointSize(qRound(fnt.pointSize()*15.0/10.0));
    m_ui->instruction->setFont(fnt);

//     m_ui->transformLabel->setFont(fnt);
//     m_ui->attribute->setFont(fnt);
    for (int i = 0; i < AnimationEffect::Generic; ++i)
        m_ui->attribute->addItem(attributeString[i]);

#define ADD_WINDOW_TYPE(_TYPE_) m_ui->windowTypes->addItem(i18n(#_TYPE_)/*, NET::_TYPE_*/);
    // Unknown
    ADD_WINDOW_TYPE(Normal);
    ADD_WINDOW_TYPE(Desktop);
    ADD_WINDOW_TYPE(Dock);
    ADD_WINDOW_TYPE(Toolbar);
    ADD_WINDOW_TYPE(Menu);
    ADD_WINDOW_TYPE(Dialog);
    ADD_WINDOW_TYPE(Override);
    ADD_WINDOW_TYPE(TopMenu);
    ADD_WINDOW_TYPE(Utility);
    ADD_WINDOW_TYPE(Splash);
    ADD_WINDOW_TYPE(DropdownMenu);
    ADD_WINDOW_TYPE(PopupMenu);
    ADD_WINDOW_TYPE(Tooltip);
    ADD_WINDOW_TYPE(Notification);
    ADD_WINDOW_TYPE(ComboBox);
    ADD_WINDOW_TYPE(DNDIcon);
#undef ADD_WINDOW_TYPE

#define ADD_EASING_CURVE(_S_, _C_) m_ui->shape->addItem(i18n(_S_), QEasingCurve::_C_)
    ADD_EASING_CURVE("Linear", Linear);

    ADD_EASING_CURVE("Accelerating (slight)", InQuad);
    ADD_EASING_CURVE("Accelerating (weak)", InCubic);
    ADD_EASING_CURVE("Accelerating (medium)", InQuart);
    ADD_EASING_CURVE("Accelerating (strong)", InQuint);
    ADD_EASING_CURVE("Run in", InBack);
    ADD_EASING_CURVE("Boing-boingg, jump", InBounce);
    ADD_EASING_CURVE("Multiple run ins", InElastic);

    ADD_EASING_CURVE("Decelerating (slight)", OutQuad);
    ADD_EASING_CURVE("Decelerating (weak)", OutCubic);
    ADD_EASING_CURVE("Decelerating (medium)", OutQuart);
    ADD_EASING_CURVE("Decelerating (strong)", OutQuint);
    ADD_EASING_CURVE("Overshoot", OutBack);
    ADD_EASING_CURVE("Fall, boingg-boing", OutBounce);
    ADD_EASING_CURVE("Multiple overshooting", OutElastic);

    ADD_EASING_CURVE("Fast in the middle (slight)", InOutQuad);
    ADD_EASING_CURVE("Fast in the middle (weak)", InOutCubic);
    ADD_EASING_CURVE("Fast in the middle (medium)", InOutQuart);
    ADD_EASING_CURVE("Fast in the middle (strong)", InOutQuint);
    ADD_EASING_CURVE("Run in and overshoot", InOutBack);

    ADD_EASING_CURVE("Slow in the middle (slight)", OutInQuad);
    ADD_EASING_CURVE("Slow in the middle (weak)", OutInCubic);
    ADD_EASING_CURVE("Slow in the middle (medium)", OutInQuart);
    ADD_EASING_CURVE("Slow in the middle (strong)", OutInQuint);
    ADD_EASING_CURVE("Overshoot and run in in the middle", OutInBack);

#if 0
    // nearly *Quad
    ADD_EASING_CURVE(InSine);
    ADD_EASING_CURVE(OutSine);
    ADD_EASING_CURVE(InOutSine);
    ADD_EASING_CURVE(OutInSine);
    // nearly *Quart - well, not exactly ;-)
    ADD_EASING_CURVE(InCirc);
    ADD_EASING_CURVE(OutCirc);
    ADD_EASING_CURVE(InOutCirc);
    ADD_EASING_CURVE(OutInCirc);
    // nearly *Quint
    ADD_EASING_CURVE(InExpo);
    ADD_EASING_CURVE(OutExpo);
    ADD_EASING_CURVE(InOutExpo);
    ADD_EASING_CURVE(OutInExpo);
#endif

#if 0
    // little idea on how to describe or what to use for...
    ADD_EASING_CURVE(InOutElastic);
    ADD_EASING_CURVE(OutInElastic);
    ADD_EASING_CURVE(InOutBounce);
    ADD_EASING_CURVE(OutInBounce);
#endif

    m_ui->shape->addItem(i18n("Bump"), 46);
#undef ADD_EASING_CURVE

    // The anchors - humans suck.
#define SET_ANCHOR(_RADIO_, _ANCHOR_) m_ui->_RADIO_->setProperty("Anchor", _ANCHOR_);
    SET_ANCHOR(saTL, AnimationEffect::Top|AnimationEffect::Left);
    SET_ANCHOR(saTM, AnimationEffect::Top);
    SET_ANCHOR(saTR, AnimationEffect::Top|AnimationEffect::Right);
    SET_ANCHOR(saML, AnimationEffect::Left);
    SET_ANCHOR(saMM, 0);
    SET_ANCHOR(saMR, AnimationEffect::Right);
    SET_ANCHOR(saBL, AnimationEffect::Bottom|AnimationEffect::Left);
    SET_ANCHOR(saBM, AnimationEffect::Bottom);
    SET_ANCHOR(saBR, AnimationEffect::Bottom|AnimationEffect::Right);
    SET_ANCHOR(saMouse, AnimationEffect::Mouse);
    SET_ANCHOR(taTL, AnimationEffect::Top|AnimationEffect::Left);
    SET_ANCHOR(taTM, AnimationEffect::Top);
    SET_ANCHOR(taTR, AnimationEffect::Top|AnimationEffect::Right);
    SET_ANCHOR(taML, AnimationEffect::Left);
    SET_ANCHOR(taMM, 0);
    SET_ANCHOR(taMR, AnimationEffect::Right);
    SET_ANCHOR(taBL, AnimationEffect::Bottom|AnimationEffect::Left);
    SET_ANCHOR(taBM, AnimationEffect::Bottom);
    SET_ANCHOR(taBR, AnimationEffect::Bottom|AnimationEffect::Right);
    SET_ANCHOR(taMouse, AnimationEffect::Mouse);
#undef SET_ANCHOR

    resize(640,480);
    adjustSize();

    connect(m_ui->animateOpacity, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(m_ui->animateMaximize, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(m_ui->slidingPopups, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(m_ui->animations, SIGNAL(currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)), SLOT(animationChanged(QTreeWidgetItem*,QTreeWidgetItem*)));
//     connect(m_ui->animations, SIGNAL(editorDestroyed(QObject*)), SLOT(itemRenamed()));
    connect(m_ui->addAnimation, SIGNAL(clicked()), SLOT(addAnimation()));
    connect(m_ui->removeAnimation, SIGNAL(clicked()), SLOT(removeAnimation()));
    connect(m_ui->importAnimations, SIGNAL(clicked()), SLOT(importAnimations()));
    connect(m_ui->exportAnimations, SIGNAL(clicked()), SLOT(exportAnimations()));

    connect(m_ui->attribute, SIGNAL(currentIndexChanged(int)), SLOT(attributeChanged(int)));
    connect(m_ui->source, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(m_ui->target, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(m_ui->useSourceX, SIGNAL(toggled(bool)), SLOT(parameterActivated(bool)));
    connect(m_ui->sourceX, SIGNAL(valueChanged(double)), SLOT(changed()));
    connect(m_ui->sXIsRelative, SIGNAL(toggled(bool)), SLOT(relativityChanged()));
    connect(m_ui->useSourceY, SIGNAL(toggled(bool)), SLOT(parameterActivated(bool)));
    connect(m_ui->sourceY, SIGNAL(valueChanged(double)), SLOT(changed()));
    connect(m_ui->sYIsRelative, SIGNAL(toggled(bool)), SLOT(relativityChanged()));
    connect(m_ui->useTargetX, SIGNAL(toggled(bool)), SLOT(parameterActivated(bool)));
    connect(m_ui->targetX, SIGNAL(valueChanged(double)), SLOT(changed()));
    connect(m_ui->tXIsRelative, SIGNAL(toggled(bool)), SLOT(relativityChanged()));
    connect(m_ui->useTargetY, SIGNAL(toggled(bool)), SLOT(parameterActivated(bool)));
    connect(m_ui->targetY, SIGNAL(valueChanged(double)), SLOT(changed()));
    connect(m_ui->tYIsRelative, SIGNAL(toggled(bool)), SLOT(relativityChanged()));

    QList<QRadioButton*> anchors = m_ui->sourceAnchor->findChildren<QRadioButton*>();
    foreach (QRadioButton *anchor, anchors)
        connect(anchor, SIGNAL(toggled(bool)), SLOT(changed()));

    anchors = m_ui->targetAnchor->findChildren<QRadioButton*>();
    foreach (QRadioButton *anchor, anchors)
        connect(anchor, SIGNAL(toggled(bool)), SLOT(changed()));

    connect(m_ui->duration, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(m_ui->shape, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(m_ui->delay, SIGNAL(valueChanged(int)), SLOT(changed()));

    m_ui->attribute->setCurrentIndex(-1);
    animationChanged(0, 0);
}

void GenericAnimationsConfig::addAnimation()
{
    QList<QTreeWidgetItem *> sel = m_ui->animations->selectedItems();
    if (sel.isEmpty()) {
        QTreeWidgetItem *item = new QTreeWidgetItem(QStringList() << i18n("New alias"));
        item->setFlags(item->flags() | Qt::ItemIsEditable);
        m_ui->animations->addTopLevelItem(item);
        m_ui->animations->editItem(item);
    }
    else {
        QTreeWidgetItem *item = new QTreeWidgetItem(QStringList() << i18n("New Animation..."));
        QTreeWidgetItem *parent = 0;
        if (!(parent = sel.at(0)->parent()))
            parent = sel.at(0);
        parent->addChild(item);
        const QString key = parent->data(0, Qt::UserRole + 2).toString();
        AniMap::iterator animation = m_animations.find(key);
        if (animation == m_animations.end())
            animation = m_animations.insert(key, QList<AniData>() << AniData());
        else
            (*animation) << AniData();
        item->setData(0, Qt::UserRole + 1, qVariantFromValue((void*)&(animation->last())));
        m_ui->animations->setCurrentItem(item);
    }
}

void GenericAnimationsConfig::removeAnimation()
{
    QList<QTreeWidgetItem *> sel = m_ui->animations->selectedItems();
    if (sel.isEmpty())
        return;

    QTreeWidgetItem *item = sel.at(0);

    if (item->parent()) { // TODO: check key and allow removing aliases
        if (AniData *data = (AniData*) item->data(0, Qt::UserRole + 1).value<void*>()) {
            QList<AniData> *list = 0;
            foreach (const QList<AniData> &animations, m_animations)
            foreach (const AniData &animation, animations)
                if (&animation == data) {
                    list = const_cast<QList<AniData>*>(&animations);
                    break;
                }

            if (list) {
                for (QList<AniData>::iterator it = list->begin(),
                                              end = list->end(); it != end; ++it) {
                    if (&(*it) == data) {
                        list->erase(it);
                        break;
                    }
                }
            }
        }
        item->setData(0, Qt::UserRole + 1, 0); // this is important to prevent crashes in ::uiToAniData
        item->parent()->removeChild(item);
        delete sel.at(0);
    }
}

void GenericAnimationsConfig::exportAnimations()
{
    QList<QTreeWidgetItem *> sel = m_ui->animations->selectedItems();
    if (sel.isEmpty())
        return;

    QTreeWidgetItem *item = sel.at(0);
    QTreeWidgetItem *parent = item->parent() ? item->parent() : item;
    if (AniData *data = (AniData*) item->data(0, Qt::UserRole + 1).value<void*>())
        uiToAniData(*data);

    const QString key = parent->data(0, Qt::UserRole + 2).toString();
    AniMap::const_iterator animations = m_animations.find(key);
    QStringList list;
    foreach (const AniData &data, *animations)
        list << data.toString();
    KMessageBox::information(0, i18n("<h3>This is an encoded string for all %1 animations that you can post or save somewhere.</h3>", parent->text(0)) + list.join(";"), i18n("Encoded %1 animations", parent->text(0)), "", KMessageBox::PlainCaption);
}

void GenericAnimationsConfig::importAnimations()
{
    QList<QTreeWidgetItem *> sel = m_ui->animations->selectedItems();
    if (sel.isEmpty())
        return;

    QTreeWidgetItem *item = sel.at(0);
    QTreeWidgetItem *parent = item->parent() ? item->parent() : item;
    const QString key = parent->data(0, Qt::UserRole + 2).toString();
    bool reallyImport = false;
    QString imported = KInputDialog::getText(i18n("Import encoded animation for %1", parent->text(0)),
                                             i18n("<h1>BE WARNED</h1>All animations for <h2>%1</h2> will be replaced!<h3>Do NOT enter some junk here</h3>", parent->text(0)),
                                             QString(), &reallyImport);
    if (reallyImport) {
        QList<QTreeWidgetItem*> kids = parent->takeChildren();
        foreach (QTreeWidgetItem *kid, kids)
            delete kid;
        AniMap::iterator animations = m_animations.find(key);
        *animations = AniData::list(imported);
        foreach (const AniData &data, *animations) {
            item = new QTreeWidgetItem(QStringList() << attributeString[data.attribute]);
            item->setData(0, Qt::UserRole + 1, qVariantFromValue((void*)&data));
            parent->addChild( item );
        }
    }
}

void GenericAnimationsConfig::animationChanged( QTreeWidgetItem *current, QTreeWidgetItem *previous )
{
    if (previous)
    if (AniData *data = (AniData*) previous->data(0, Qt::UserRole + 1).value<void*>()) {
        uiToAniData(*data);
        previous->setText(0, attributeString[m_ui->attribute->currentIndex()]);
    }

    int page = 1;
    if (current) {
        if (AniData *data = (AniData*)current->data(0, Qt::UserRole + 1).value<void*>()) {
            page = 0;
            m_ui->removeAnimation->setEnabled(true);
            aniDataToUi(*data);
        }
        else
            m_ui->removeAnimation->setEnabled(false); //TODO: check key
        QString title;
        if (current->parent())
            title = current->parent()->text(0) + " / ";
        title += current->text(0);
        m_ui->currentAnimation->setText(title);
//         m_ui->animationSettingsGroup->setTitle(title);
    }
    else
        m_ui->removeAnimation->setEnabled(false);

    m_ui->animationSettings->setCurrentIndex(page);
}

void GenericAnimationsConfig::setSuffix(const QString &str)
{
    m_ui->sourceX->setSuffix(m_ui->sXIsRelative->isChecked() ? " %" : str);
    m_ui->sourceY->setSuffix(m_ui->sYIsRelative->isChecked() ? " %" : str);
    m_ui->targetX->setSuffix(m_ui->tXIsRelative->isChecked() ? " %" : str);
    m_ui->targetY->setSuffix(m_ui->tYIsRelative->isChecked() ? " %" : str);
}

void GenericAnimationsConfig::setRange(float min, float max)
{
    m_ui->sourceX->setRange(min, max);
    m_ui->sourceY->setRange(min, max);
    m_ui->targetX->setRange(min, max);
    m_ui->targetY->setRange(min, max);
}

void GenericAnimationsConfig::attributeChanged(int idx)
{
    changed();
    bool jump = false;
    switch (idx) {
    case AnimationEffect::Opacity:
        setRange(0.0,1.0);
        jump = true;
    case AnimationEffect::Brightness:
    case AnimationEffect::Saturation:
        m_ui->sourceAxis->hide();
        m_ui->targetAxis->hide();
        if (!jump)
            setRange(0.0,2.0);
        setSuffix("    ");
        jump = true;
    case AnimationEffect::Rotation:
        if (!jump) {
            setRange(-360.0,360.0);
            m_ui->sourceAnchor->show();
            m_ui->targetAnchor->show();
            m_ui->sourceAxis->show();
            m_ui->targetAxis->show();
            setSuffix(QString::fromUtf8("°  "));
        } else {
            m_ui->sourceAnchor->hide();
            m_ui->targetAnchor->hide();
        }
        m_ui->sXIsRelative->hide();
        m_ui->sYIsRelative->hide();
        m_ui->tXIsRelative->hide();
        m_ui->tYIsRelative->hide();
        m_ui->sourceY->hide();
        m_ui->targetY->hide();
        m_ui->useSourceX->hide();
        m_ui->useSourceX->setChecked(true);
        m_ui->useTargetX->hide();
        m_ui->useTargetX->setChecked(true);
        m_ui->useSourceY->hide();
        m_ui->useTargetY->hide();
        m_ui->sourceX->setPrefix("");
        m_ui->targetX->setPrefix("");
        break;
    case AnimationEffect::Position:
    case AnimationEffect::Size:
        setRange(0.0,5000.0);
        setSuffix(" px");
        jump = true;
    case AnimationEffect::Translation:
        if (!jump) {
            setRange(-5000.0,5000.0);
            setSuffix(" px");
            jump = true;
        }
    case AnimationEffect::Scale: {
        if (!jump) {
            setRange(0.0,4.0);
            setSuffix("    ");
        }
        m_ui->useSourceX->show();
        m_ui->useTargetX->show();
        m_ui->useSourceY->show();
        m_ui->useTargetY->show();
        m_ui->sourceY->show();
        m_ui->targetY->show();
        m_ui->sourceX->setPrefix("x: ");
        m_ui->targetX->setPrefix("x: ");
        m_ui->sXIsRelative->show();
        m_ui->sYIsRelative->show();
        m_ui->tXIsRelative->show();
        m_ui->tYIsRelative->show();
        const QString relation = (idx == AnimationEffect::Translation) ?
                                                i18nc("follows a spinbox 'n %'", "of window") :
                                                i18nc("follows a spinbox 'n %'", "of screen");
        m_ui->sXIsRelative->setText(relation);
        m_ui->sYIsRelative->setText(relation);
        m_ui->tXIsRelative->setText(relation);
        m_ui->tYIsRelative->setText(relation);
        m_ui->sourceAnchor->setVisible(idx != AnimationEffect::Translation);
        m_ui->targetAnchor->setVisible(idx != AnimationEffect::Translation);
        m_ui->sourceAxis->hide();
        m_ui->targetAxis->hide();
        break;
    }
    case AnimationEffect::Clip: {
            setRange(0.0,1.0);
            setSuffix("    ");
            m_ui->useSourceX->show();
            m_ui->useTargetX->show();
            m_ui->useSourceY->show();
            m_ui->useTargetY->show();
            m_ui->sourceY->show();
            m_ui->targetY->show();
            m_ui->sourceX->setPrefix("x: ");
            m_ui->targetX->setPrefix("x: ");
            m_ui->sXIsRelative->hide();
            m_ui->sYIsRelative->hide();
            m_ui->tXIsRelative->hide();
            m_ui->tYIsRelative->hide();
            m_ui->sourceAnchor->setVisible(true);
            m_ui->targetAnchor->setVisible(true);
            m_ui->sourceAxis->hide();
            m_ui->targetAxis->hide();
            break;
    }
    default: break;
    }
}

void GenericAnimationsConfig::defaults()
{
    // DO NOTHING!
    // "defaults" would mean to jsut delete all user settings, the UI can do this pretty fast and
    // such action would be to invasive and irreversible.
}

#define PARSE(_KEY_,_LABEL_) \
animations = m_animations.insert(_KEY_, AniData::list(conf.readEntry(_KEY_, defaultAnimations.value(_KEY_, QString())))); \
item = new QTreeWidgetItem(QStringList() << i18n(_LABEL_)); \
item->setData(0, Qt::UserRole + 2, _KEY_);\
foreach (const AniData &data, *animations) {\
    anItem = new QTreeWidgetItem(QStringList() << attributeString[data.attribute]);\
    anItem->setData(0, Qt::UserRole + 1, qVariantFromValue((void*)&data));\
    item->addChild( anItem ); \
}\
m_ui->animations->addTopLevelItem( item )

void GenericAnimationsConfig::load()
{
    KCModule::load();
    KConfigGroup conf = effects->effectConfig( "GenericAnimations" );

    m_ui->animateOpacity->setChecked(conf.readEntry("Opacity", true));
    m_ui->animateMaximize->setChecked(conf.readEntry("UnMaximize", true));
    m_ui->slidingPopups->setChecked(conf.readEntry("SlidingPopups", true));

    QTreeWidgetItem *item, *anItem;
    AniMap::iterator animations;
    PARSE("ActivationAnimations", "Activation");
    PARSE("ImplicitActivationAnimations", "Implicit Activations");
    PARSE("DeactivationAnimations", "Deactivation");
    PARSE("AddAnimations", "Show");
    PARSE("CloseAnimations", "Close");
    PARSE("MinimizeAnimations", "Minimize");
    PARSE("UnminimizeAnimations", "Unminimize");
    PARSE("LeavingDesktopAnimations", "Leaving Current Desktop");
    PARSE("EnteringDesktopAnimations", "Entering Current Desktop");
    PARSE("LeavingDesktopReverseAnimations", "Leaving Current Desktop (backwards)");
    PARSE("EnteringDesktopReverseAnimations", "Entering Current Desktop (backwards)");
    PARSE("ShowDesktopAnimations", "ShowDesktop");
    PARSE("HideDesktopAnimations", "HideDesktop");
    if (!m_ui->animations->currentItem())
        m_ui->animations->setCurrentItem( m_ui->animations->topLevelItem(0) );
    emit changed(false);
}


int GenericAnimationsConfig::metaData( MetaType type, uint meta ) const
{
    switch (type) {
        case SourceAnchor:
            return ((meta>>5) & 0x1f);
        case TargetAnchor:
            return (meta& 0x1f);
        case RelativeSourceX:
        case RelativeSourceY:
        case RelativeTargetX:
        case RelativeTargetY: {
            const int shift = 10 + type - RelativeSourceX;
            return ((meta>>shift) & 1);
        }
        case Axis:
            return ((meta>>10) & 3);
        default:
            return 0;
    }
}

void GenericAnimationsConfig::setMetaData( MetaType type, uint value, uint &meta ) const
{
    switch (type) {
        case SourceAnchor:
            meta &= ~(0x1f<<5);
            meta |= ((value & 0x1f)<<5);
            break;
        case TargetAnchor:
            meta &= ~(0x1f);
            meta |= (value & 0x1f);
            break;
        case RelativeSourceX:
        case RelativeSourceY:
        case RelativeTargetX:
        case RelativeTargetY: {
            const int shift = 10 + type - RelativeSourceX;
            if (value)
                meta |= (1<<shift);
            else
                meta &= ~(1<<shift);
            break;
        }
        case Axis:
            meta &= ~(3<<10);
            meta |= ((value & 3)<<10);
            break;
        default:
            break;
    }
}

void GenericAnimationsConfig::aniDataToUi(const AniData &d)
{
    for (int i = 0; i < m_ui->windowTypes->count(); ++i)
        m_ui->windowTypes->item(i)->setSelected(d.windowType & (1<<i));

    m_ui->attribute->setCurrentIndex(d.attribute);

    uint anchor;
    QList<QRadioButton*> buttons;
    if (d.from.isValid()) {
        m_ui->useSourceX->setChecked(   d.attribute == AnimationEffect::Translation ||
                                        d.attribute == AnimationEffect::Rotation ||
                                        d.from[0] != defaultValue(d.attribute));
        m_ui->sXIsRelative->setChecked(metaData( RelativeSourceX, d.meta ));
        m_ui->sourceX->setValue(m_ui->sXIsRelative->isChecked() ? d.from[0] * 100 : d.from[0]);
        m_ui->useSourceY->setChecked(   d.attribute == AnimationEffect::Translation ||
                                        d.attribute == AnimationEffect::Rotation ||
                                        d.from[1] != defaultValue(d.attribute));
        m_ui->sYIsRelative->setChecked(metaData( RelativeSourceY, d.meta ));
        m_ui->sourceY->setValue(m_ui->sYIsRelative->isChecked() ? d.from[1] * 100 : d.from[1]);
        m_ui->sourceAxis->setCurrentIndex(metaData( Axis, d.meta ));

        anchor = metaData( SourceAnchor, d.meta );
        buttons = m_ui->sourceAnchor->findChildren<QRadioButton*>();
        foreach (QRadioButton *button, buttons)
            button->setChecked(button->property("Anchor") == anchor);
    }

    if (d.to.isValid()) {
        m_ui->useTargetX->setChecked(   d.attribute == AnimationEffect::Translation ||
                                        d.attribute == AnimationEffect::Rotation ||
                                        d.to[0] != defaultValue(d.attribute));
        m_ui->tXIsRelative->setChecked(metaData( RelativeTargetX, d.meta ));
        m_ui->targetX->setValue(m_ui->tXIsRelative->isChecked() ? d.to[0] * 100 : d.to[0]);
        m_ui->useTargetY->setChecked(   d.attribute == AnimationEffect::Translation ||
                                        d.attribute == AnimationEffect::Rotation ||
                                        d.to[1] != defaultValue(d.attribute));
        m_ui->tYIsRelative->setChecked(metaData( RelativeTargetY, d.meta ));
        m_ui->targetY->setValue(m_ui->tYIsRelative->isChecked() ? d.to[1] * 100 : d.to[1]);
        m_ui->targetAxis->setCurrentIndex(metaData( Axis, d.meta ));

        anchor = metaData( TargetAnchor, d.meta );
        buttons = m_ui->targetAnchor->findChildren<QRadioButton*>();
        foreach (QRadioButton *button, buttons)
            button->setChecked(button->property("Anchor") == anchor);
    }

    m_ui->source->setChecked(d.from.isValid());
    m_ui->target->setChecked(d.to.isValid());

    m_ui->duration->setValue(d.duration);
    m_ui->shape->setCurrentIndex(m_ui->shape->findData(d.customCurve));
    m_ui->delay->setValue(d.time);
}

float GenericAnimationsConfig::defaultValue(AnimationEffect::Attribute a) const
{
    if (a == AnimationEffect::Rotation || a == AnimationEffect::Translation)
        return 0.0f;
    if (a < AnimationEffect::NonFloatBase)
        return 1.0f;
    return -1.0f;
}

void GenericAnimationsConfig::uiToAniData(AniData &d)
{
    int netWmTypeMask = 0;
    for (int i = 0; i < m_ui->windowTypes->count(); ++i)
        if (m_ui->windowTypes->item(i)->isSelected())
            netWmTypeMask |= (1<<i);
    d.windowType = (NET::WindowTypeMask)netWmTypeMask;

    d.attribute = (AnimationEffect::Attribute)m_ui->attribute->currentIndex();

    uint anchor(0);
    QList<QRadioButton*> buttons;
    if (m_ui->source->isChecked()) {
        double value[2] = { m_ui->sourceX->value(), m_ui->sourceY->value() };
        const bool relative[2] = {
            m_ui->sXIsRelative->isVisibleTo(m_ui->source) && m_ui->sXIsRelative->isChecked(),
            m_ui->sYIsRelative->isVisibleTo(m_ui->source) && m_ui->sYIsRelative->isChecked() };
        if (relative[0]) value[0] /= 100.0;
        if (relative[1]) value[1] /= 100.0;
        d.from.set(m_ui->useSourceX->isChecked() ? value[0] : defaultValue(d.attribute),
                   m_ui->useSourceY->isChecked() ? value[1] : defaultValue(d.attribute));

        setMetaData( RelativeSourceX, relative[0], d.meta );
        setMetaData( RelativeSourceY, relative[1], d.meta );

        if (m_ui->sourceAnchor->isVisibleTo(m_ui->source)) {
            buttons = m_ui->sourceAnchor->findChildren<QRadioButton*>();
            foreach (QRadioButton *button, buttons)
                if (button->isChecked()) {
                    anchor = button->property("Anchor").toUInt();
                    break;
                }
                setMetaData( SourceAnchor, anchor, d.meta );
        }

        if (m_ui->sourceAxis->isVisibleTo(m_ui->source))
            setMetaData( Axis, (uint)m_ui->sourceAxis->currentIndex(), d.meta );
    }
    else
        d.from.invalidate();

    if (m_ui->target->isChecked()) {
        double value[2] = { m_ui->targetX->value(), m_ui->targetY->value() };
        const bool relative[2] = {
            m_ui->tXIsRelative->isVisibleTo(m_ui->target) && m_ui->tXIsRelative->isChecked(),
            m_ui->tYIsRelative->isVisibleTo(m_ui->target) && m_ui->tYIsRelative->isChecked() };
        if (relative[0]) value[0] /= 100.0;
        if (relative[1]) value[1] /= 100.0;
        d.to.set(m_ui->useTargetX->isChecked() ? value[0] : defaultValue(d.attribute),
                 m_ui->useTargetY->isChecked() ? value[1] : defaultValue(d.attribute));

        setMetaData( RelativeTargetX, relative[0], d.meta );
        setMetaData( RelativeTargetY, relative[1], d.meta );

        if (m_ui->targetAnchor->isVisibleTo(m_ui->target)) {
            buttons = m_ui->targetAnchor->findChildren<QRadioButton*>();
            foreach (QRadioButton *button, buttons)
                if (button->isChecked()) {
                    anchor = button->property("Anchor").toUInt();
                    break;
                }
                setMetaData( TargetAnchor, anchor, d.meta );
        }
        if (m_ui->targetAxis->isVisibleTo(m_ui->target))
            setMetaData( Axis, (uint)m_ui->targetAxis->currentIndex(), d.meta );
    }
    else
        d.to.invalidate();

    d.duration = m_ui->duration->value();
    d.customCurve = m_ui->shape->itemData(m_ui->shape->currentIndex()).toInt();
    d.time = m_ui->delay->value();
}

void GenericAnimationsConfig::parameterActivated(bool active)
{
    QDoubleSpinBox *box_0 = 0;
    QCheckBox *box_1 = 0;
    if (sender() == m_ui->useSourceX) {
        box_0 = m_ui->sourceX;
        box_1 = m_ui->sXIsRelative;
    } else if (sender() == m_ui->useSourceY) {
        box_0 = m_ui->sourceY;
        box_1 = m_ui->sYIsRelative;
    } else if (sender() == m_ui->useTargetX) {
        box_0 = m_ui->targetX;
        box_1 = m_ui->tXIsRelative;
    } else if (sender() == m_ui->useTargetY) {
        box_0 = m_ui->targetY;
        box_1 = m_ui->tYIsRelative;
    }
    if (box_0) {
        box_0->setEnabled(active);
        box_1->setEnabled(active);
        QGroupBox *gb = static_cast<QGroupBox*>(box_0->parentWidget());
        if (!gb->isChecked()) {
            gb->setChecked(true);
            gb->setChecked(false);
        }
    }
}

void GenericAnimationsConfig::relativityChanged()
{
    QDoubleSpinBox *box = 0;
    if (sender() == m_ui->sXIsRelative)
        box = m_ui->sourceX;
    else if (sender() == m_ui->sYIsRelative)
        box = m_ui->sourceY;
    else if (sender() == m_ui->tXIsRelative)
        box = m_ui->targetX;
    else if (sender() == m_ui->tYIsRelative)
        box = m_ui->targetY;
    if (box) {
        if (static_cast<QCheckBox*>(sender())->isChecked())
            box->setSuffix(" %");
        else {
            switch (m_ui->attribute->currentIndex()) {
            case AnimationEffect::Size:
            case AnimationEffect::Translation:
                box->setSuffix(" px");
                break;
            default:
                box->setSuffix("    ");
            }
        }
        changed();
    }
}

#define STORE(_KEY_) \
list.clear();\
animations = m_animations.find(_KEY_); \
foreach (const AniData &data, *animations)\
    list << data.toString();\
conf.writeEntry(_KEY_, list.join(";"))


void GenericAnimationsConfig::save()
{
    QList<QTreeWidgetItem *> sel = m_ui->animations->selectedItems();
    if (!sel.isEmpty()) {
        if (AniData *data = (AniData*) sel.at(0)->data(0, Qt::UserRole + 1).value<void*>()) {
            uiToAniData(*data);
            sel.at(0)->setText(0, attributeString[m_ui->attribute->currentIndex()]);
        }
    }

    QStringList list;
    AniMap::const_iterator animations;
    KConfigGroup conf = EffectsHandler::effectConfig("GenericAnimations");
    conf.writeEntry("Opacity", m_ui->animateOpacity->isChecked());
    conf.writeEntry("UnMaximize", m_ui->animateMaximize->isChecked());
    conf.writeEntry("SlidingPopups", m_ui->slidingPopups->isChecked());

    STORE("ActivationAnimations");
    STORE("ImplicitActivationAnimations");
    STORE("DeactivationAnimations");
    STORE("AddAnimations");
    STORE("CloseAnimations");
    STORE("MinimizeAnimations");
    STORE("UnminimizeAnimations");
    STORE("LeavingDesktopAnimations");
    STORE("EnteringDesktopAnimations");
    STORE("LeavingDesktopReverseAnimations");
    STORE("EnteringDesktopReverseAnimations");
    STORE("ShowDesktopAnimations");
    STORE("HideDesktopAnimations");

    conf.sync();
    emit changed(false);
    EffectsHandler::sendReloadMessage( "generic_animations" );
}

#include "generic_animations_config.moc"
