/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2009 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#ifndef KWIN_SNOW_CONFIG_H
#define KWIN_SNOW_CONFIG_H

#include <kcmodule.h>

#include "ui_generic_animations_config.h"
#if HAVE_KWIN_4_8
#include <kwinanimationeffect.h>
#else
#include "animationeffect.h"
#endif

class KActionCollection;

namespace KWin
{
class AniData;
class GenericAnimationsConfigForm : public QWidget, public Ui::GenericAnimationsConfigForm
{
    Q_OBJECT
public:
    explicit GenericAnimationsConfigForm(QWidget* parent);
};

class GenericAnimationsConfig : public KCModule
{
    Q_OBJECT
public:
    explicit GenericAnimationsConfig(QWidget* parent = 0, const QVariantList& args = QVariantList());

public slots:
    void save();
    void load();
    void defaults();
private:
    void aniDataToUi(const AniData &d);
    float defaultValue(AnimationEffect::Attribute a) const;
    void setSuffix(const QString &);
    void setRange(float, float);
    void uiToAniData(AniData &d);
private slots:
    void addAnimation();
    void animationChanged(QTreeWidgetItem *, QTreeWidgetItem *);
    void attributeChanged(int idx);
    void exportAnimations();
    void importAnimations();
    void parameterActivated(bool active);
    void removeAnimation();
    void relativityChanged();
private:
    GenericAnimationsConfigForm* m_ui;
    typedef QMap<QString, QList<AniData> > AniMap;
    AniMap m_animations;
    QString attributeString[AnimationEffect::Generic];
private: //TODO replace in KWin 4.8 by lib calls
    enum MetaType { SourceAnchor, TargetAnchor,
                    RelativeSourceX, RelativeSourceY, RelativeTargetX, RelativeTargetY, Axis };
    int metaData(MetaType type, uint meta ) const;
    void setMetaData(MetaType type, uint value, uint &meta ) const;
    //END TODO ====================
};

} // namespace

#endif
