/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

 Copyright (C) 2010 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "befaded.h"
// #include <kwinconfig.h>
#include <kconfiggroup.h>
// #include <cmath>

using namespace KWin;

KWIN_EFFECT( befaded, BeFaded )

BeFaded::BeFaded()
{
    m_tabboxActive = 0;
    theOldStackingOrder = effects->stackingOrder();
    theJustMapped = 0;
    reconfigure( ReconfigureAll );
    connect (effects, SIGNAL(windowAdded(KWin::EffectWindow*)), SLOT(windowAdded(KWin::EffectWindow*)));
    connect (effects, SIGNAL(windowUnminimized(KWin::EffectWindow*)), SLOT(windowUnminimized(KWin::EffectWindow*)));
    connect (effects, SIGNAL(windowDeleted(KWin::EffectWindow*)), SLOT(windowDeleted(KWin::EffectWindow*)));
    connect (effects, SIGNAL(stackingOrderChanged()), SLOT(stackingOrderChanged()));
    connect (effects, SIGNAL(tabBoxAdded(int)), SLOT(tabBoxAdded()));
    connect (effects, SIGNAL(tabBoxClosed()), SLOT(tabBoxClosed()));
}

void BeFaded::reconfigure( ReconfigureFlags )
{
//     KConfigGroup conf = effects->effectConfig( "BeFaded" );
//     myDuration = conf.readEntry( "AnimationDuration", 280 );
}

void BeFaded::stackingOrderChanged()
{
    if ( effects->activeFullScreenEffect() || m_tabboxActive )
    {   // implicitly shared copy is faster than a perhaps scanning comparism ;-)
        theOldStackingOrder = effects->stackingOrder();
        return;
    }

    EffectWindowList newStackingOrder = effects->stackingOrder();
    if (newStackingOrder != theOldStackingOrder)
    {
        if ((theUpmostWindow = upmost(newStackingOrder)))
        {
            if (theUpmostWindow == theJustMapped )
                theJustMapped = 0;
            else if (theUpmostWindow != upmost(theOldStackingOrder))
                windowRaised(theUpmostWindow);
        }
        theOldStackingOrder = newStackingOrder;
    }
}

void BeFaded::tabBoxAdded()
{
    ++m_tabboxActive;
}

void BeFaded::tabBoxClosed()
{
    m_tabboxActive = qMax(--m_tabboxActive, 0);
}

EffectWindow *BeFaded::upmost( const EffectWindowList &stack )
{
    if (stack.isEmpty())
        return 0L;
    EffectWindowList::const_iterator i = stack.constEnd();
    while ( i != stack.constBegin() )
    {
        --i;
        EffectWindow *w = *i;
        if (!w)
            continue;
        if ( w->isDeleted() || w->isMinimized() || w->isTooltip() || w->isPopupMenu() || w->isComboBox() ||
             w->isDropdownMenu() || w->isMenu() || w->keepAbove() || w->isDNDIcon() ||
             w->isNotification() || w->isSplash() ||
             w->isDock() || w->isToolbar() || w->isDesktop() )
            continue;
        return w;
    }
    return 0L;
}

void BeFaded::windowAdded( EffectWindow* w )
{
    theJustMapped = w;
}
 
void BeFaded::windowDeleted( EffectWindow* w )
{
    theOldStackingOrder.removeAll( w );
    if (w == theUpmostWindow)
        theUpmostWindow = 0;
    if (w == theJustMapped)
        theJustMapped = 0;
}

void BeFaded::windowRaised( EffectWindow *w )
{
    int i = theOldStackingOrder.indexOf(w);
    if (i < 0)
        return;
    bool fadedIn = false;
    while (++i < theOldStackingOrder.count()) {
        if (EffectWindow *sibling = theOldStackingOrder.at(i)) {
            if (sibling->isOnDesktop(w->desktop()) || sibling->isOnAllDesktops()) { // TODO: activities
                const QRect sg = sibling->geometry(), wg = w->geometry();
                // the new top was entirely covered by this one, don't fade it but the risen one in
                if (sg.contains(wg)) {
                    if (!fadedIn)
                        animate(w, Opacity, 0, 250, FPx2(1.0), QEasingCurve::OutCubic, 0, FPx2(0.0));
                    fadedIn = true;
                } else if (sg.intersects(wg)) {
                    // can't Bump because we also have to elevate the window while it fades out
                    effects->setElevatedWindow(sibling, true);
                    animate(sibling, Opacity, 1, 200, FPx2(0.0), QEasingCurve::OutCubic, 0, FPx2(1.0));
                }
            }
        }
    }
}

void BeFaded::animationEnded(EffectWindow *w, Attribute a, uint meta)
{
    if (meta == 1 && a == Opacity) { // this is teh fade-out of a sibling, bring it back
        effects->setElevatedWindow(w, false);
        animate(w, Opacity, 0, 300, FPx2(1.0), QEasingCurve::InCubic, 0, FPx2(0.0));
    }
}

void BeFaded::windowUnminimized( EffectWindow* w )
{
    theJustMapped = w;
}

